# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""File for holding constants used in reviewcheck."""
import os
from pathlib import Path
from typing import List


class Constants:
    """Class for holding constants used in reviewcheck."""

    COLORS: List[str] = [
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
        "bright_black",
        "bright_green",
        "bright_yellow",
        "bright_blue",
        "bright_magenta",
        "bright_cyan",
        "bright_white",
        "dark_cyan",
        "turquoise4",
        "spring_green1",
        "slate_blue3",
        "slate_blue3",
        "light_pink4",
        "medium_purple3",
        "light_slate_blue",
        "dark_goldenrod",
        "dark_olive_green3",
        "hot_pink3",
        "thistle3",
        "khaki1",
        "indian_red",
    ]

    CONFIG_DIR: Path = Path(os.environ.get("XDG_CONFIG", Path.home() / ".config"))
    CONFIG_PATH: Path = CONFIG_DIR / "reviewcheckrc"

    CACHE_DIR: Path = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache"))
    DATA_DIR: Path = CACHE_DIR / "reviewcheck"
    COMMENT_NOTE_IDS_PATH: Path = DATA_DIR / "old_comment_ids"

    TUI_AUTHOR_WIDTH = 16
    TUI_DATE_WIDTH = 12
    TUI_TWO_COL_PADDING_WIDTH = 7
    TUI_THREE_COL_PADDING_WIDTH = 10

    THREADPOOL_MAXSIZE = 32



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""File containing utility functions."""
import json
import logging
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import requests

from reviewcheck.constants import Constants
from reviewcheck.exceptions import RCException


class Utils:
    """Class that contains utility functions for reviewcheck."""

    @staticmethod
    def convert_time(timestamp: str) -> str:
        """Convert GitLab timestamps to human-readable timestamps.

        Convert timestamps from the format that GitLab uses to a human
        readable one like so:

            <date> <short month> <hour><minute>
        """
        try:
            time = datetime.strptime(
                timestamp,
                "%Y-%m-%dT%H:%M:%S.%f%z",
            )
        except ValueError:
            try:
                time = datetime.strptime(
                    timestamp[0:-6],
                    "%Y-%m-%dT%H:%M:%S.%f",
                )
            except ValueError:
                raise RCException(f"Couldn't parse GitLab timestamp '{timestamp}'")
        human_readable_time = datetime.strftime(
            time,
            "%d %b %H:%M",
        )
        return human_readable_time

    @staticmethod
    def download_data(
        params: Tuple[str, str, Optional[str], Any]
    ) -> Tuple[Any, Any, Any]:
        """Download data for MR, and for reaction if requested.

        This function just calls download_gitlab_data() twice. The point
        of this is that it makes it possible to download both reaction
        data and merge request data in the same loop in a multithreading
        executor.
        """
        token, mr_url, reaction_url, metadata = params
        reaction_response = []
        if reaction_url:
            reaction_response = Utils.download_gitlab_data(token, reaction_url)
        mr_response = Utils.download_gitlab_data(token, mr_url)
        return mr_response, reaction_response, metadata

    @staticmethod
    def download_gitlab_data(
        secret_token: str,
        url: str,
    ) -> List[Dict[str, Any]]:
        """Download each page of data for a given GitLab MR.

        Download the data for a given URL. All pages are downloaded and
        then combined to one list. The data mainly contains the comments
        and threads on the MR with associated metadata.

        :param secret_token: Token to access the GitLab API.
        :param url: URLs to download data from.
        :return: All data for the MR as a list.
        """
        # This is how to create a reusable connection pool with python
        # requests.

        with requests.Session() as session:
            session.mount(
                "https://",
                requests.adapters.HTTPAdapter(
                    pool_maxsize=Constants.THREADPOOL_MAXSIZE,
                    max_retries=3,
                    pool_block=True,
                ),
            )
            response = session.get(url, headers={"PRIVATE-TOKEN": secret_token})
            response_json = json.loads(response.content)
            logging.info(
                "request was completed in %s seconds [%s]",
                response.elapsed.total_seconds(),
                response.url,
            )
            if response.status_code != 200:
                logging.error(
                    "request failed, error code %s [%s]",
                    response.status_code,
                    response.url,
                )
            if 500 <= response.status_code < 600:
                # server is overloaded? give it a break
                time.sleep(5)

            num_pages = int(response.headers["X-Total-Pages"])
            for page in range(2, num_pages + 1):
                response = session.get(
                    f"{url}&page={page}", headers={"PRIVATE-TOKEN": secret_token}
                )
                response_json += json.loads(response.content)
                logging.info(
                    "request was completed in %s seconds [%s]",
                    response.elapsed.total_seconds(),
                    response.url,
                )
                if response.status_code != 200:
                    logging.error(
                        "request failed, error code %s [%s]",
                        response.status_code,
                        response.url,
                    )
                if 500 <= response.status_code < 600:
                    # server is overloaded? give it a break
                    time.sleep(5)

        if isinstance(response_json, list):
            if len(response_json) == 0 or isinstance(response_json[0], dict):
                return response_json

        raise Exception("Malformed data returned from GitLab.")



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""File for storing the MergeRequest class."""
import re
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional

from reviewcheck.exceptions import RCException


class MergeRequest:
    """Class representing a merge request."""

    def __init__(self, threads: Any, reactions: Any, metadata: Any, user: str):
        """Initialize a MergeRequest object."""
        if reactions is None:
            reactions = []
        if (
            not isinstance(threads, list)
            or not isinstance(reactions, list)
            or not isinstance(metadata, dict)
        ):
            raise RCException("Malformed data from GitHub")
        self.title: str = metadata["title"]
        self.mr_author: str = metadata["author"]["username"]
        self.user: str = user
        self.is_author = self.mr_author == self.user
        self.id: int = metadata["iid"]
        self.project: int = metadata["project_id"]
        self.web_url: str = metadata["web_url"]
        self.upvotes: int = metadata["upvotes"]
        self.creation_time: str = metadata["created_at"]
        self.source_branch: str = metadata["source_branch"]
        self.description: str = metadata["description"]
        self.jira_ticket_number = self.extract_jira()

        self.threads: List[Dict[str, Any]] = []
        self.number_of_open_threads = 0
        self.number_of_open_threads_for_user = 0
        self.number_of_open_threads_needing_user_reply = 0
        self.all_last_message_ids: List[str] = []
        for thread in threads:
            messages = thread["notes"]
            first_message = thread["notes"][0]
            last_message = thread["notes"][-1]
            # Filter out comments that are not threads (not resolvable)
            if "resolved" in first_message:
                # ...and only those that are not already resolved
                if not first_message["resolved"]:
                    self.number_of_open_threads += 1
                    # ...and the thread is relevant for the user
                    if (
                        self.user_is_participant(messages)
                        or self.user_is_referenced_in_thread(messages)
                        or self.user_is_author_of_thread(messages)
                    ):
                        self.number_of_open_threads_for_user += 1
                        self.threads.append(thread)
                        self.all_last_message_ids.append(last_message["id"])
                        if last_message["author"]["username"] != self.user:
                            self.number_of_open_threads_needing_user_reply += 1

        self.reaction_and_name: DefaultDict[str, List[str]] = defaultdict(list)
        self.reaction_and_gitlab_user: DefaultDict[str, List[str]] = defaultdict(list)
        for reaction in reactions:
            self.reaction_and_name[reaction["name"]].append(reaction["user"]["name"])
            self.reaction_and_gitlab_user[reaction["name"]].append(
                reaction["user"]["username"]
            )

    def jira_link(self, jira_base_url: str) -> Optional[str]:
        """Getter for JIRA URL."""
        if self.jira_ticket_number:
            return f"{jira_base_url}/{self.jira_ticket_number}"
        return "No JIRA reference found"

    def extract_jira(self) -> Optional[str]:
        """Extract JIRA ticket number from MR description.

        :param description: The description of the MR.
        :return: Jira ticket number if found, otherwise None.
        """
        if self.description is None:
            return None
        # Parse the VIRA ticket number
        jira_regex = re.compile(r".*(?i)JIRA: (.*)(\\n)*")
        jira_match = jira_regex.match(self.description.split("\n")[-1])
        jira = None
        if jira_match:
            jira = str(jira_match.group(1))
            already_a_link_regex = re.compile(r"\[(.*)\].*")
            already_a_link = already_a_link_regex.match(jira)
            if already_a_link:
                jira = str(already_a_link.group(1))
        return jira

    def print_reactors(self) -> Optional[str]:
        """:return: Printable list of people who have reacted."""
        all_reactors: List[str] = list(
            set(
                person
                for reaction in self.reaction_and_name.values()
                for person in reaction
            )
        )
        all_reactors.sort()
        if len(all_reactors) > 0:
            return " | ".join(all_reactors)
        return None

    def print_upvoters(self) -> Optional[str]:
        """:return: Printable list of people who have upvoted."""
        all_upvoters = self.reaction_and_name.get("thumbsup", [])
        all_upvoters.sort()
        return " | ".join(all_upvoters)

    def user_reacted_but_no_upvote(self) -> bool:
        """Return True if user has left a reaction without upvoting.

        :return: True if the user forgot to thumb up, otherwise False.
        """
        has_reacted = any(
            self.user in names for names in self.reaction_and_gitlab_user.values()
        )
        has_upvoted = self.user in self.reaction_and_gitlab_user.get("thumbsup", [])
        if has_reacted and not has_upvoted:
            return True
        return False

    def user_is_reviewer(self) -> bool:
        """Return true if the user is reviewing this MR.

        The user is considered to be reviewing the merge request when
        they have left a reaction on it but not upvoted it, and they
        themself is not the author of the merge request because they
        cannot review their own merge request.
        """
        return not (self.user_reacted_but_no_upvote() and not self.is_author)

    def user_is_referenced_in_thread(self, messages: List[Dict[str, Any]]) -> bool:
        """Check if configured user is mentioned in a given thread.

        :param messages: The thread to check as a list of messages.
        :param username: The username of the configured user.

        :return: True if the user is mentioned in the given thread,
            otherwise false.
        """
        for message in messages:
            if ("@" + self.user) in message["body"]:
                return True
        return False

    def user_is_author_of_thread(self, messages: List[Dict[str, Any]]) -> bool:
        """Check if the given user has responded to thread in its MR.

        Check whether the given user is the author of the current merge
        request, and if so, whether is the last person to respond to the
        given thread.

        :param mr: The merge request to check.
        :param messages: The thread to check.
        :param username: The username to look for as author.

        :return: False if the user doesn't have anything to respond to,
            otherwise True.
        """
        if self.mr_author == self.user:
            if messages[-1]["author"]["username"] != self.user:
                return True
        return False

    def user_is_participant(self, messages: List[Dict[str, Any]]) -> bool:
        """Check if user is active in current thread.

        :param messages: The thread to check whether the user is active
            in.
        :param: The username of the user to check for involvement of.

        :return: True if the user is involved in the current thread,
            otherwise False.
        """
        for message in messages:
            if message["author"]["username"] == self.user:
                return True
        return False



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Functions for generating components with the rich module."""
from typing import Any, Dict, List

from rich import box
from rich.table import Table

from reviewcheck.constants import Constants
from reviewcheck.merge_request import MergeRequest
from reviewcheck.utils import Utils


class RichGenerator:
    """Static functions for generatign rich text components."""

    @staticmethod
    def thread_table(
        row_styles: List[str],
        border_color: str,
        main_color: str,
        width: int,
    ) -> Table:
        """Get an empty threads table to populate with messages."""
        table = Table(
            show_header=True,
            show_lines=True,
            row_styles=row_styles,
            border_style=border_color,
            header_style=f"bold {main_color}",
            width=width,
            box=box.ROUNDED,
        )

        table.add_column("Date", width=Constants.TUI_DATE_WIDTH)
        table.add_column("Author", width=Constants.TUI_AUTHOR_WIDTH)
        table.add_column(
            "Message",
            style="dim",
            min_width=width
            - Constants.TUI_AUTHOR_WIDTH
            - Constants.TUI_DATE_WIDTH
            - Constants.TUI_THREE_COL_PADDING_WIDTH,
        )
        return table

    @staticmethod
    def rows_highlighting(
        comment: Dict[str, Any],
        needs_reply: bool,
        username: str,
    ) -> List[str]:
        """Return highligh configuration for each note in a thread.

        Messages after the user's own last message should be
        highlighted. This marks them as "new". It is assumed that all
        messages that have been added after the user's own last message
        is new to the user.

        :param comment: The thread to provide highlighting for.
        :param needs_reply: Whether there are new messages after the
            user's last message.
        :param username: Username of the user who needs to reply.

        :return: List of highlighting information to be given to the
            rich library function for highlighting the given thread.
        """
        n_replies = len(comment["notes"])
        notes = comment["notes"]
        author_list = [n["author"]["username"] for n in notes]
        if not needs_reply:
            return [""] * (n_replies)
        elif username not in author_list:
            return ["bold white not dim"] * (n_replies)
        i_my_last_reply = n_replies - author_list[::-1].index(username)
        return [""] * (i_my_last_reply) + ["bold white not dim"] * (
            n_replies - i_my_last_reply
        )

    @staticmethod
    def info_box_title(
        mr: MergeRequest,
        color: str,
    ) -> str:
        """Return the info box title with rich text configuration.

        :param mr: Data about the mr to construct title for.
        :param jira_ticket_number: The JIRA ticket number associated
            with the MR.
        :param color: Color to use for the title when printing.

        :return: The formatted title to use for the info box.
        """
        title_elements = [
            f"[bold {color}]{mr.title}",
            f"!{mr.id}",
        ]

        if mr.jira_ticket_number:
            title_elements.append(mr.jira_ticket_number)

        return " | ".join(title_elements)

    @staticmethod
    def info_box_content(
        mr: MergeRequest,
        jira_base_url: str,
    ) -> str:
        """Return the content of the info box for an merge request.

        :param mr: Data for the merge request in question.
        :param jira_url: URL to JIRA ticket associated with the merge
            request, or replacement text.
        :param jira: JIRA ticket number.
            the user is involved.
        :param n_response_required: Number of threads on the merge
            request where the user is involved and hasn't replied to a
            message.

        :return: The text to put in the info box for the given merge
            request.
        """
        info = f"Upvotes: {mr.upvotes}"
        if mr.upvotes > 0:
            info += f"\nPeople who have upvoted: {mr.print_upvoters()}"
        reactors = mr.print_reactors()
        if reactors:
            info += f"\nPeople who have reacted: {reactors}"
        info += f"\nOpen discussions: {mr.number_of_open_threads}"
        if mr.number_of_open_threads:
            info += (
                f"\nOpen discussions where you are involved: "
                f"{mr.number_of_open_threads_for_user}"
            )
            info += "\nOpen discussions you need to respond (colored border): "
            info += f"{mr.number_of_open_threads_needing_user_reply}"
        info += f"\n\nGitLab link:   {mr.web_url}"
        info += f"\nJira link:     {mr.jira_link(jira_base_url)}"
        info += f"\nSource branch: {mr.source_branch}"
        info += f"\nCreated at:    {Utils.convert_time(mr.creation_time)}"
        info += f"\n\n{mr.description}"
        return info



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Parse the command line arguments given to reviewcheck."""
import argparse
from argparse import Namespace, RawTextHelpFormatter

import shtab


class Cli:
    """Class with the functions associated with argument parsing."""

    @staticmethod
    def check_positive_int(value: str) -> int:
        """Return positive int if possible, otherwise raise exception.

        Validator for checking that a variable is a positive integer and
        converting it to the int type.

        :param value: The value given on command line.

        :raises:ArgumentTypeError: Raised when the value cannot be
            converted to a positive integer.

        :return: The value as a positive int if possible.
        """
        try:
            ivalue = int(value)
        except Exception:
            raise argparse.ArgumentTypeError("Flag argument must be a positive integer")

        if ivalue > 0:
            return ivalue
        else:
            raise argparse.ArgumentTypeError("Flag argument must be a positive integer")

    @staticmethod
    def parse_arguments() -> Namespace:
        """Parse the arguments given on the command line.

        Create an argument parser and use it to parse the arguments
        given on the command line.

        :return: The parsed arguments as a Namespace object.
        """
        parser = argparse.ArgumentParser(
            description="""
        This script will tell you which discussions on GitLab you need to respond to.
        You will be shown the discussions from MRs where:

        - You are the author
        - Someone has replied to your comment but you haven't replied back
        - Someone has referenced your @username

        The script needs to be configured before you can use it. Add a
        ~/.config/reviewcheckrc with the following content:

        ```
        secret_token: <get a secret token from your settings in gitlab>
        user: <your username>
        ```

        There is also some optional configuration:

        ```
        ignored_mrs: [123, 546, 789] (defult: empty list)
        show_all_discussions: true (defult: false)
        hide_replied_discussions: true (default: false)
        ```
        """,
            formatter_class=RawTextHelpFormatter,
        )

        parser.add_argument(
            "--version",
            help="Show version",
            dest="print_version",
            required=False,
            action="store_true",
            default=False,
        )

        shtab.add_argument_to(parser, ["-s", "--print-completion"])

        parser.add_argument(
            "-a",
            "--all",
            help="Show all threads, even when you don't need to reply",
            action="store_true",
            default=False,
            dest="all",
        )

        parser.add_argument(
            "-u",
            "--user",
            help="Username whose reviews you want to analyze",
            action="store",
            default=False,
            dest="user",
        )

        parser.add_argument(
            "-i",
            "--ignore",
            help="Space separated list of MRs to ignore, e.g. '-i 373 371'",
            action="store",
            nargs="+",
            default=[],
            dest="ignore",
        )

        parser.add_argument(
            "-r",
            "--refresh",
            help="If set, data will be refreshed at an interval specified in minutes",
            type=Cli.check_positive_int,
            action="store",
            default=None,
            dest="refresh_time",
        )

        parser.add_argument(
            "-m",
            "--minimal",
            help="Only show discussion where a reply is needed.",
            action="store_true",
            default=False,
            dest="minimal",
        )

        parser.add_argument(
            "-w",
            "--width",
            help="Terminal display width",
            type=Cli.check_positive_int,
            action="store",
            dest="output_width",
        )

        parser.add_argument(
            "-N",
            "--no-notifications",
            help="Do not invoke notify-send(1)",
            action="store_true",
            default=False,
        )

        subparsers = parser.add_subparsers(dest="command")

        subparsers.add_parser(
            "configure",
            help="Interactively set up configuration file",
            description="Asks for input to write to the configuration file.",
        )

        return parser.parse_args()



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""File containing the Config class for working with config files."""
from typing import Any, Dict, Optional

import yaml

from reviewcheck.constants import Constants


class Config:
    """Class for interacting with the configuration file."""

    def __init__(self, reconfigure: bool):
        """Set initial variable state for the Config class.

        :param reconfigure: Whether or not the user has requested to
            update the configuration of reviewcheck.
        """
        self.reconfigure = reconfigure

    def setup_configuration(self) -> None:
        """Let the user configure the application.

        This will be done if the user has requested to configure
        reviewcheck or if no config file can be found.
        """
        # Only attempt to write the configuration file if it does not
        # already exist.
        if self.reconfigure or not Constants.CONFIG_PATH.exists():
            Constants.CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)

            if not self.reconfigure:
                print(
                    f"No configuration file found at {str(Constants.CONFIG_PATH)}, "
                    "please provide some information to populate it:"
                )

            token = input("GitLab API token: ")
            username = input("Username: ")
            api_url = input("API URL: ")
            jira_url = input("Jira URL: ")
            project_ids = input("Project IDs (space-separated): ")

            if project_ids:
                project_ids_list = [int(i) for i in project_ids.split(" ")]
            else:
                project_ids_list = []

            config_object: Dict[str, Any] = {
                "secret_token": token,
                "user": username,
                "api_url": api_url,
                "jira_url": jira_url,
                "project_ids": project_ids_list,
            }

            with open(Constants.CONFIG_PATH, "w") as f:
                f.write(yaml.safe_dump(config_object))

    def get_configuration(self) -> Optional[Dict[str, Any]]:
        """Read the configuration file into a dictionary.

        Reads the configuration file into a dictionary. If the
        configuration file does not exist, queries the user for
        information and writes it.

        :return: A dict containing the contents of the configuration
            file, or None if the file does not exist.
        """
        self.setup_configuration()
        try:
            with open(Constants.CONFIG_PATH, "r") as f:
                content = yaml.safe_load(f)
        except FileNotFoundError:
            return None

        if isinstance(content, dict):
            return content
        else:
            print(
                f"ERROR - the configuration {Constants.CONFIG_PATH} couldn't"
                "be parsed as a dictionary."
            )

        return None



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Script to show what threads in GitLab you've forgotten to respond to.

You have to configure the script before running it by running
reviewcheck --configure.
"""
import json
import logging
import subprocess
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from shutil import get_terminal_size
from typing import Any, Dict, List, Set

import requests
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress
from rich.text import Text

from reviewcheck.cli import Cli
from reviewcheck.config import Config
from reviewcheck.constants import Constants
from reviewcheck.exceptions import RCException
from reviewcheck.merge_request import MergeRequest
from reviewcheck.rich_components import RichGenerator
from reviewcheck.utils import Utils

console = Console()


def configure() -> int:
    """Set up the configuration of reviewcheck.

    :return: Always returns 0.
    """
    Config(True).setup_configuration()
    return 0


def read_viewed_message_ids() -> Set[str]:
    """Read stored comment note IDs from last check.

    :return: The IDs of those comments that were considered in need of a
        reply on the last run of reviewcheck.
    """
    old_comment_note_ids: Set[str] = set()
    try:
        with open(Constants.COMMENT_NOTE_IDS_PATH) as f:
            lines = f.readlines()
    except FileNotFoundError:
        return old_comment_note_ids

    for line in lines:
        old_comment_note_ids.add(line.strip())

    return old_comment_note_ids


def write_viewed_message_ids(new_comment_note_ids: Set[str]) -> None:
    """Write all the comment note IDs from the last run to file.

    :param new_comment_note_ids: The IDs of the comments that are
        considered in need of a reply during this run of reviewcheck.
    """
    with open(Constants.COMMENT_NOTE_IDS_PATH, "w") as f:
        for id in new_comment_note_ids:
            f.write(f"{id}\n")


def show_reviews(config: Dict[str, Any], suppress_notifications: bool) -> None:
    """Download MR data and present review info for each relevant MR.

    :param config: The resolved configuration of reviewcheck.
    """
    secret_token = config["secret_token"]
    api_url = config["api_url"]
    jira_url = config["jira_url"]
    project_ids = config["project_ids"]
    user = config["user"]
    ignored_mrs = config["ignored_mrs"]
    show_all_discussions = config["show_all_discussions"]
    hide_all_threads_user_already_replied_to = config["hide_replied_discussions"]

    mr_pages = []
    with Progress(transient=True, expand=True) as progress:
        gitlab_download_task = progress.add_task(
            "[green]Downloading MR data...",
            start=False,
        )
        for project in project_ids:
            api_url_project = f"{api_url}/projects/{project}"
            merge_requests_data = json.loads(
                requests.get(
                    f"{api_url_project}/merge_requests?state=opened&per_page=500",
                    headers={"PRIVATE-TOKEN": secret_token},
                ).content
            )
            if isinstance(merge_requests_data, dict):
                raise Exception("Malformed data received from GitLab")
            else:
                mr_pages += [
                    mr
                    for mr in merge_requests_data
                    if str(mr["iid"]) not in ignored_mrs
                ]

        mrs: List[MergeRequest] = []

        def reaction_url(project: str, id: str) -> str:
            """Construct API URL for merge request reactions."""
            return f"{api_url}/projects/{project}/merge_requests/{id}/award_emoji"

        def mr_url(project: str, id: str) -> str:
            """Construct API URL for merge request notes."""
            return (
                f"{api_url}/projects/{project}/merge_requests/{id}"
                "/discussions?per_page=500"
            )

        with ThreadPoolExecutor(max_workers=Constants.THREADPOOL_MAXSIZE) as executor:
            progress.start_task(gitlab_download_task)
            progress.update(gitlab_download_task, total=len(mrs))

            for mr_response, reaction_response, mr in executor.map(
                Utils.download_data,
                [
                    (
                        secret_token,
                        mr_url(mr["project_id"], mr["iid"]),
                        reaction_url(mr["project_id"], mr["iid"]),
                        mr,
                    )
                    for mr in mr_pages
                ],
            ):
                mrs.append(
                    MergeRequest(
                        mr_response,
                        reaction_response,
                        mr,
                        user,
                    )
                )
                progress.update(gitlab_download_task, advance=1)

    console.print(
        Panel(
            Text(
                f"Status as of {datetime.now().strftime('%Y-%m-%d %H:%M')}",
                justify="center",
            ),
            style="reverse bold",
        ),
        width=config["output_width"],
    )

    ids_of_seen_messages = read_viewed_message_ids()
    for mr in mrs:
        if (
            mr.user_is_reviewer()
            and mr.number_of_open_threads_needing_user_reply == 0
            and (not show_all_discussions or len(mr.threads) == 0)
        ):
            continue

        main_mr_color = Constants.COLORS[mr.id % len(Constants.COLORS)]
        mr_info_header = Panel(
            Text(RichGenerator.info_box_content(mr, jira_url)),
            title=RichGenerator.info_box_title(mr, main_mr_color),
            width=config["output_width"],
        )

        console.print(mr_info_header)
        if (
            mr.user_reacted_but_no_upvote()
            and not mr.is_author
            and mr.number_of_open_threads_needing_user_reply == 0
        ):
            continue

        for thread in mr.threads:
            last_message = thread["notes"][-1]

            # When minimal view is requsted, only show threads where
            # a response is required.
            if hide_all_threads_user_already_replied_to:
                if last_message["author"]["username"] == user:
                    continue

            user_needs_to_reply = False
            if last_message["author"]["username"] != user:
                user_needs_to_reply = True

                if (
                    not suppress_notifications
                    and str(last_message["id"]) not in ids_of_seen_messages
                ):
                    subprocess.run(
                        [
                            "notify-send",
                            "--expire-time=15000",
                            f"New comment on MR !{mr.id}",
                            (
                                f"{last_message['author']['name']} writes:\n\n"
                                f"{last_message['body']}"
                            ),
                        ]
                    )

            border_color = f"{main_mr_color}" if user_needs_to_reply else "white"
            row_highlighting_style = RichGenerator.rows_highlighting(
                thread,
                user_needs_to_reply,
                user,
            )
            thread_table = RichGenerator.thread_table(
                row_styles=row_highlighting_style,
                border_color=border_color,
                main_color=main_mr_color,
                width=config["output_width"],
            )
            for message in thread["notes"]:
                update_time = Utils.convert_time(message["updated_at"])
                thread_table.add_row(
                    update_time,
                    message["author"]["name"],
                    message["body"],
                )

            if user_needs_to_reply or show_all_discussions:
                thread_table.add_row(
                    "",
                    "Discussion link",
                    f"{mr.web_url}#note_{thread['notes'][0]['id']}",
                )

            console.print(thread_table)

    write_viewed_message_ids(set([id for mr in mrs for id in mr.all_last_message_ids]))


def run() -> int:
    """Start execution of reviewcheck."""
    args = Cli.parse_arguments()

    if args.print_version:
        from reviewcheck import __version__

        print(__version__)
        return 0

    command_palette = {
        "configure": configure,
    }

    if args.command:
        func = command_palette.get(args.command)

        if func is not None:
            return func()
        else:
            print(
                f"'{args.command} is not a valid command, please see "
                "`reviewcheck --help`."
            )
            return 127

    Constants.DATA_DIR.mkdir(exist_ok=True)

    config = Config(False).get_configuration()

    if config is None:
        print(f"Could not read configuration from {str(Constants.CONFIG_PATH)}.")
        return 1

    if args.user:
        config["user"] = args.user
    config["user"] = config["user"].upper()

    if args.output_width:
        config["output_width"] = args.output_width
    else:
        config.setdefault("output_width", get_terminal_size().columns)

    if "ignored_mrs" not in config:
        config["ignored_mrs"] = args.ignore

    if "show_all_discussions" not in config:
        config["show_all_discussions"] = args.all

    if "hide_replied_discussions" not in config:
        config["hide_replied_discussions"] = args.minimal

    try:
        if args.refresh_time is None:
            show_reviews(config, args.no_notifications)
            return 0

        while True:
            console.clear()
            show_reviews(config, args.no_notifications)
            time.sleep(args.refresh_time * 60)
    except KeyboardInterrupt:
        print("\nBye bye!")
        return 0
    except RCException as e:
        logging.error(f"Reviewcheck encountered a problem: {e}")
        return 1



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""File containing custom reviewcheck exceptions."""


class RCException(Exception):
    """General exception for hiding stack trace in reviewcheck.

    Base exception to inherit from when creating custom exceptions and
    to use directly when it's not necessary to differentiate excepitons.
    It is caught in main and will print the message in the exceptions.
    """



# Copyright 2023 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Main file for reviewcheck, where execution starts."""
import sys

from reviewcheck import app

if __name__ == "__main__":
    sys.exit(app.run())
